/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Copyright (c) 2005 Arnaud Patard <arnaud.patard@rtp-net.org>
 * Samsung S3C2410 keyboard support
 *
 * Based on various pxa ipaq drivers.
 * 
 * ChangeLog
 *
 * 2006-07-30: RAG for LN2410SBC
 *      - Added check for timer pending
 *
 * 2005-07-24: Arnaud Patard <arnaud.patard@rtp-net.org>
 * 	- Added key repeat
 *
 * 2005-07-23: Arnaud Patard <arnaud.patard@rtp-net.org>
 * 	- Renamed all kbd occurences to something like buttons
 * 	- Added debounce (specially usefull for iPAQ power button)
 *
 * 2005-06-21: Arnaud Patard <arnaud.patard@rtp-net.org>
 *      - Initial version
 *
 */

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <asm/arch/regs-gpio.h>
#include <asm/arch/buttons.h>

/* For id.version */
#define S3C2410BUTVERSION	0x0001
#define DRV_NAME		"s3c2410-buttons"
#define DRV_VERSION	0x0001

#define REPEAT_DELAY		HZ/10

//#define BUTTON_DEBUG

#ifdef BUTTON_DEBUG
  #define dprintk(msg...) printk("s3c2410_buttons: " msg);
#else
  #define dprintk(msg...)
#endif

MODULE_AUTHOR("Arnaud Patard <arnaud.patard@rtp-net.org>");
MODULE_DESCRIPTION("s3c2410 buttons driver");
MODULE_LICENSE("GPL");



struct s3c2410_buttons_private {
	struct device		*parent;
	struct input_dev	*dev;
	spinlock_t		lock;
	int 			count;
	int			shift;	
};

static struct s3c2410_buttons_private priv;

int mod_loaded = 0;

#define ENABLE_SYSFS
#ifdef ENABLE_SYSFS
static ssize_t ln2410_evup_show(struct device *dev,struct device_attribute *attr,char *buf)
{
        int temp,i;
        struct platform_device *pdev;
	struct s3c2410_button *s3c2410_buttons;
	struct s3c2410_button *bp;
	struct s3c2410_butt_mach_info *butt_info;
        pdev = to_platform_device(dev);
	butt_info = pdev->dev.platform_data;
	s3c2410_buttons = butt_info->buttons;
        temp = 0;
        for (i = 0, bp = s3c2410_buttons; i < butt_info->size;bp++, i++) 
	{
            temp |= (bp->ev_up)<<i;         
	}
	return snprintf(buf,PAGE_SIZE,"%d \n",temp);
}
static ssize_t ln2410_evup_store(struct device *dev,struct device_attribute *attr,const char *buf,size_t count)
{

	int temp,i;
        struct platform_device *pdev;
	struct s3c2410_button *s3c2410_buttons;
	struct s3c2410_button *bp;
	struct s3c2410_butt_mach_info *butt_info;
        
	pdev = to_platform_device(dev);
	butt_info = pdev->dev.platform_data;
	s3c2410_buttons = butt_info->buttons;
        if(sscanf(buf,"%d",&temp) != 1)
	{
		return -EINVAL;
	}
        for (i = 0, bp = s3c2410_buttons; i < butt_info->size; bp++, i++) 
	{
            bp->ev_up = (temp>>i)&0x01;         
	}
	return strnlen(buf,count);
}
static DEVICE_ATTR(evup,0664,ln2410_evup_show,ln2410_evup_store);
static ssize_t ln2410_evdown_show(struct device *dev,struct device_attribute *attr,char *buf)
{
        int temp,i;
        struct platform_device *pdev;
	struct s3c2410_button *s3c2410_buttons;
	struct s3c2410_button *bp;
	struct s3c2410_butt_mach_info *butt_info;

        temp = 0;
        pdev = to_platform_device(dev);
	butt_info = pdev->dev.platform_data;
	s3c2410_buttons = butt_info->buttons;
        for (i = 0, bp = s3c2410_buttons; i < butt_info->size; bp++, i++) 
	{
            temp |= (bp->ev_down)<<i;         
	}
	return snprintf(buf,PAGE_SIZE,"%d \n",temp);
}
static ssize_t ln2410_evdown_store(struct device *dev,struct device_attribute *attr,const char *buf,size_t count)
{

	int temp,i;
        struct platform_device *pdev;
	struct s3c2410_button *s3c2410_buttons;
	struct s3c2410_button *bp;
	struct s3c2410_butt_mach_info *butt_info;

        pdev = to_platform_device(dev);
	butt_info = pdev->dev.platform_data;
	s3c2410_buttons = butt_info->buttons;
        if(sscanf(buf,"%d",&temp) != 1)
	{
		return -EINVAL;
	}
        for (i = 0, bp = s3c2410_buttons; i < butt_info->size; bp++, i++) 
	{
            bp->ev_down = (temp>>i)&0x01;         
	}
	return strnlen(buf,count);
}
static DEVICE_ATTR(evdown,0664,ln2410_evdown_show,ln2410_evdown_store);
static ssize_t ln2410_evrepeat_show(struct device *dev,struct device_attribute *attr,char *buf)
{
        int temp,i;

        struct platform_device *pdev;
	struct s3c2410_button *s3c2410_buttons;
	struct s3c2410_button *bp;
	struct s3c2410_butt_mach_info *butt_info;

        temp = 0;
        pdev = to_platform_device(dev);
	butt_info = pdev->dev.platform_data;
	s3c2410_buttons = butt_info->buttons;
        for (i = 0, bp =s3c2410_buttons; i < butt_info->size;bp++, i++) 
	{
            temp |= (bp->ev_repeat)<<i;         
	}
	return snprintf(buf,PAGE_SIZE,"%d \n",temp);
}
static ssize_t ln2410_evrepeat_store(struct device *dev,struct device_attribute *attr,const char *buf,size_t count)
{

	int temp,i;
        struct platform_device *pdev;
	struct s3c2410_button *s3c2410_buttons;
	struct s3c2410_button *bp;
	struct s3c2410_butt_mach_info *butt_info;

        pdev = to_platform_device(dev);
	butt_info = pdev->dev.platform_data;
	s3c2410_buttons = butt_info->buttons;
        if(sscanf(buf,"%d",&temp) != 1)
	{
		return -EINVAL;
	}
        for (i = 0, bp = s3c2410_buttons; i < butt_info->size; bp++, i++) 
	{
            bp->ev_repeat = (temp>>i)&0x01;         
	}
	return strnlen(buf,count);
}
static DEVICE_ATTR(evrepeat,0664,ln2410_evrepeat_show,ln2410_evrepeat_store);

static void reg_button_attributes(struct platform_device *pdev)
{
	device_create_file(&(pdev->dev),&dev_attr_evup);
	device_create_file(&(pdev->dev),&dev_attr_evdown);
	device_create_file(&(pdev->dev),&dev_attr_evrepeat);
}

static void remove_button_attributes(struct platform_device *pdev)
{
	device_remove_file(&(pdev->dev),&dev_attr_evup);
	device_remove_file(&(pdev->dev),&dev_attr_evdown);
	device_remove_file(&(pdev->dev),&dev_attr_evrepeat);
}
#endif

static irqreturn_t s3c2410but_keyevent(int irq, void *dev)
{
	struct s3c2410_button *button = (struct s3c2410_button *)dev;
	int down;


	if (!button)
		return IRQ_HANDLED;
	if (!mod_loaded)
		return IRQ_HANDLED;
        if(timer_pending(&button->timer))
		return IRQ_HANDLED;
	down = !(s3c2410_gpio_getpin(button->pin));
	if (button->last_state == down)
		return IRQ_HANDLED;
	button->last_state = down;

        if(mod_loaded)
	{
	   down = !(s3c2410_gpio_getpin(button->pin));
	   button->last_state = down;
	   dprintk("%s button %s\n",button->name, down ? "pressed" : "released");
	   if(down)
	   {
               if(button->ev_down)
               {
		  input_report_key(priv.dev, button->keycode, down);
	   	  input_sync(priv.dev);
               }
           }
	   else
           {
               if(button->ev_up)
               {
		  input_report_key(priv.dev, button->keycode, down);
	   	  input_sync(priv.dev);
               }
           }

	   if(&button->timer)
           {
              if (down)
		 mod_timer(&button->timer, jiffies + REPEAT_DELAY);
           }
        }

	return IRQ_HANDLED;
}

static void s3c2410but_timer_callback(unsigned long data)
{
		struct s3c2410_button *button = (struct s3c2410_button *) data;
		int down;

		down = !(s3c2410_gpio_getpin(button->pin));
		
		if (down != button->last_state) 
		{
			dprintk("Timer: %s button %s\n",button->name, down ? "pressed" : "released");
			button->last_state = down;
                        if(down)
                        {
			   if(button->ev_down)
                           {
		              input_report_key(priv.dev, button->keycode, down);
	   	              input_sync(priv.dev);
                           }
                         }
                         else
                         {
                           if(button->ev_up)
                           {
		              input_report_key(priv.dev, button->keycode, down);
	   	              input_sync(priv.dev);
                           }
                         }
		}
		else
		{
			if(down)
                        {
                           dprintk("Refiring Timer: %s button %s\n",
                                    button->name, down ? "pressed" : "released");
                           if(button->ev_repeat)
                           {
		              input_report_key(priv.dev, button->keycode, down);
	   	              input_sync(priv.dev);
                           }
			   mod_timer(&button->timer, jiffies + REPEAT_DELAY);
			}			
		}
}

static int __init s3c2410but_probe(struct platform_device *pdev)
{
	struct input_dev *input_dev;
	struct s3c2410_button *s3c2410_buttons;
	struct s3c2410_button *bp;
	struct s3c2410_butt_mach_info *butt_info;
	int i,err;

	butt_info = pdev->dev.platform_data;
	s3c2410_buttons = butt_info->buttons;
	/* Initialise input stuff */
	dprintk("Initializing s3c2410_buttons \n");
	memset(&priv, 0, sizeof(struct s3c2410_buttons_private));
	input_dev = input_allocate_device();
	if (!input_dev) {
		printk(KERN_ERR "Unable to allocate the input device !!\n");
		return -ENOMEM;
	}
   	priv.parent	     = &pdev->dev;
	priv.dev	     = input_dev;
	priv.dev->evbit[0] = BIT(EV_KEY);

	priv.dev->private    = &priv;
	priv.dev->cdev.dev   = &pdev->dev;
	priv.dev->name       = DRV_NAME;
	priv.dev->id.bustype = BUS_HOST;
	priv.dev->id.vendor  = 0xDEAD;
	priv.dev->id.product = 0xBEEF;
	priv.dev->id.version = DRV_VERSION;


	dprintk("Initializing s3c_button arrays \n");
        i = butt_info->size;
	dprintk("Array size = %d \n",i);
	for (i = 0, bp = s3c2410_buttons; i < butt_info->size; i++, bp++) {
		dprintk("i = %d \n",i);
		set_bit(bp->keycode, priv.dev->keybit);
                s3c2410_gpio_cfgpin(bp->pin, S3C2410_GPIO_SFN2);
		bp->irq = s3c2410_gpio_getirq(bp->pin);
		s3c2410_gpio_pullup(bp->pin,0);	
		err = request_irq (bp->irq, s3c2410but_keyevent,
			     IRQF_SAMPLE_RANDOM | IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			     bp->name, bp);
		if (err) {
			dev_err(&pdev->dev, "error requesting irq %d\n", bp->irq);
			goto err_irq;
		}
		init_timer(&bp->timer);
		bp->timer.function = s3c2410but_timer_callback;
		bp->timer.data     = (unsigned long)bp;
	}

	printk(KERN_INFO "%s successfully loaded\n", DRV_NAME);
	#ifdef ENABLE_SYSFS
		reg_button_attributes(pdev);
	#endif
	mod_loaded = 1;
	/* All went ok, so register to the input system */
	input_register_device(input_dev);

	return 0;
err_irq:
	for (--i,--bp ; i >= 0; i--, bp--) {
		disable_irq(bp->irq);
		free_irq(bp->irq, bp);
	}
	input_free_device(input_dev);	
	return err;
}

static int s3c2410but_remove(struct platform_device *pdev)
{
	int i;
	struct s3c2410_button *s3c2410_buttons;
	struct s3c2410_button *bp;
	struct s3c2410_butt_mach_info *butt_info;

	butt_info = pdev->dev.platform_data;
	s3c2410_buttons = butt_info->buttons;

	for (i = 0, bp = s3c2410_buttons; i < butt_info->size; i++, bp++) {
		disable_irq(bp->irq);
		free_irq(bp->irq, bp);
	}
	#ifdef ENABLE_SYSFS
		remove_button_attributes(pdev);
	#endif
	input_unregister_device(priv.dev);

	return 0;
}


static struct platform_driver s3c2410but_driver = {
	.driver		= {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
	},
       .probe          = s3c2410but_probe,
       .remove         = s3c2410but_remove,
};


int __init s3c2410but_init(void)
{
	return platform_driver_register(&s3c2410but_driver);
}

void __exit s3c2410but_exit(void)
{
	platform_driver_unregister(&s3c2410but_driver);
}

module_init(s3c2410but_init);
module_exit(s3c2410but_exit);




