/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Copyright (c) 2005 Arnaud Patard <arnaud.patard@rtp-net.org>
 * Samsung S3C2410 keyboard support
 *
 * Based on various pxa ipaq drivers.
 * 
 */

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <asm/arch/regs-gpio.h>
#include <asm/arch/buttons.h>

/* For id.version */
#define DRV_VERSION	0x0001
#define DRV_NAME	"s3c-buttons"

#define REPEAT_DELAY	HZ/10

MODULE_AUTHOR("Arnaud Patard <arnaud.patard@rtp-net.org>");
MODULE_DESCRIPTION("s3c24xx buttons driver");
MODULE_LICENSE("GPL");

struct s3c_buttons_private {
	struct device		*parent;
	struct input_dev	*dev;
	spinlock_t		lock;
	int 			count;
	int			shift;	
};

static struct s3c_buttons_private priv;

static irqreturn_t s3c_but_keyevent(int irq, void *dev)
{
	struct s3c_button *button = (struct s3c_button *)dev;
	int down;

	if (!button)
		return IRQ_HANDLED;

	down = !(s3c2410_gpio_getpin(button->pin));

	/* the power button of the ipaq are tricky. They send 'released'
	 * events even when the button are already released. The work-around
	 * is to proceed only if the state changed.
	 */
	if (button->last_state == down)
		return IRQ_HANDLED;

	button->last_state = down;
	
	dev_dbg(priv.parent, "%s button %s\n", button->name,
		down ? "pressed" : "released");

	input_report_key(priv.dev, button->keycode, down);
	input_sync(priv.dev);

	if (down)
		mod_timer(&button->timer, jiffies + REPEAT_DELAY);

	return IRQ_HANDLED;
}

static void s3c_but_timer_callback(unsigned long data)
{
		struct s3c_button *button = (struct s3c_button *) data;
		int down;

		down = !(s3c2410_gpio_getpin(button->pin));
		
		if (down) {
			dev_dbg(priv.parent, "Timer: %s button %s\n",
				button->name, down ? "pressed" : "released");
			input_report_key(priv.dev, button->keycode, down);
			input_sync(priv.dev);
			mod_timer(&button->timer, jiffies + REPEAT_DELAY);
		}
}

static int __init s3c_but_probe(struct platform_device *pdev)
{
	struct input_dev *input_dev;
	struct s3c_button *s3c_buttons;
	struct s3c_button *bp;
	struct s3c_butt_mach_info *butt_info;
	int i, err;

	butt_info = pdev->dev.platform_data;
	s3c_buttons = butt_info->buttons;

	/* Initialise input stuff */
	memset(&priv, 0, sizeof(struct s3c_buttons_private));

	input_dev = input_allocate_device();
	if (!input_dev) {
		printk(KERN_ERR "Unable to allocate the input device !!\n");
		return -ENOMEM;
	}

	priv.parent	     = &pdev->dev;
	priv.dev	     = input_dev;
	priv.dev->evbit[0]   = BIT(EV_KEY);

	priv.dev->private    = &priv;
	priv.dev->cdev.dev   = &pdev->dev;
	priv.dev->name       = DRV_NAME;
	priv.dev->id.bustype = BUS_HOST;
	priv.dev->id.vendor  = 0xDEAD;
	priv.dev->id.product = 0xBEEF;
	priv.dev->id.version = DRV_VERSION;

	for (i = 0, bp = s3c_buttons; i < butt_info->size; i++, bp++) {
		set_bit(bp->keycode, priv.dev->keybit);
		bp->irq = s3c2410_gpio_getirq(bp->pin);
		s3c2410_gpio_cfgpin(bp->pin, S3C2410_GPIO_SFN2);
		err = request_irq (bp->irq, s3c_but_keyevent,
			     IRQF_SAMPLE_RANDOM | IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			     bp->name, bp);
		if (err) {
			dev_err(&pdev->dev, "error requesting irq %d\n", bp->irq);
			goto err_irq;
		}

		if (bp->wake)
			enable_irq_wake(bp->irq);

		init_timer(&bp->timer);
		bp->timer.function = s3c_but_timer_callback;
		bp->timer.data     = (unsigned long)bp;
	}
	
	dev_info(&pdev->dev, "successfully loaded\n");

	/* All went ok, so register to the input system */
	input_register_device(priv.dev);

	return 0;

err_irq:
	for (--i,--bp ; i >= 0; i--, bp--) {
		if (bp->wake)
			disable_irq_wake(bp->irq);
		disable_irq(bp->irq);
		free_irq(bp->irq, bp);
	}
	input_free_device(input_dev);	
	return err;
}

static int s3c_but_remove(struct platform_device *pdev)
{
	int i;
	struct s3c_button *s3c_buttons;
	struct s3c_button *bp;
	struct s3c_butt_mach_info *butt_info;

	butt_info = pdev->dev.platform_data;
	s3c_buttons = butt_info->buttons;

	for (i = 0, bp = s3c_buttons; i < butt_info->size; i++, bp++) {
		if (bp->wake)
			disable_irq_wake(bp->irq);
		disable_irq(bp->irq);
		free_irq(bp->irq, bp);
	}

	input_unregister_device(priv.dev);

	return 0;
}


static struct platform_driver s3c_but_driver = {
	.driver		= {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
	},
	.probe		= s3c_but_probe,
	.remove		= s3c_but_remove,
};


static int __init s3c_but_init(void)
{
	return platform_driver_register(&s3c_but_driver);
}

static void __exit s3c_but_exit(void)
{
	platform_driver_unregister(&s3c_but_driver);
}

module_init(s3c_but_init);
module_exit(s3c_but_exit);
