/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Copyright (c) 2005 Arnaud Patard <arnaud.patard@rtp-net.org>
 * Samsung S3C2410-LP35 "joystick" support
 *
 * Based on various pxa ipaq drivers and s3c2410_buttons.
 * 
 * ChangeLog
 *
 * 2006-08-18: RAG created LP35 JS version
 * 	- This drivers minics a one button mouse using "joystick" buttons on LN2410SBC-LP35
 *
 * 2006-07-30: RAG for LN2410SBC
 *      - Added check for timer pending
 *
 * 2005-07-24: Arnaud Patard <arnaud.patard@rtp-net.org>
 * 	- Added key repeat
 *
 * 2005-07-23: Arnaud Patard <arnaud.patard@rtp-net.org>
 * 	- Renamed all kbd occurences to something like buttons
 * 	- Added debounce (specially usefull for iPAQ power button)
 *
 * 2005-06-21: Arnaud Patard <arnaud.patard@rtp-net.org>
 *      - Initial version
 *
 */


#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <asm/io.h>
#include <linux/irq.h>

#include <asm/arch/regs-gpio.h>
#include <asm/arch/js_buttons.h>

/* For id.version */
#define S3C2410JSBUTVERSION	0x0001
#define DRV_NAME		"s3c2410-LP35-js"

#define REPEAT_DELAY		HZ/10

//#define JSBUTTON_DEBUG

#ifdef JSBUTTON_DEBUG
  #define dprintk(msg...) printk("s3c2410-LP35-JS: " msg);
#else
  #define dprintk(msg...)
#endif

MODULE_AUTHOR("Arnaud Patard <arnaud.patard@rtp-net.org>");
MODULE_DESCRIPTION("s3c2410-LP35-JS driver");
MODULE_LICENSE("GPL");


/* To be moved later to a better place */
/*
static struct s3c2410_jsbutton s3c2410_jsbuttons[] = {
	{   IRQ_EINT9,  S3C2410_GPG1,    S3C2410_GPG1_EINT9,     0, "S3C2410_jsbuttons LF", 0},
	{  IRQ_EINT16,  S3C2410_GPG8,    S3C2410_GPG8_EINT16,    1, "S3C2410_jsbuttons RT", 0},
	{  IRQ_EINT17,  S3C2410_GPG9,    S3C2410_GPG9_EINT17,    2, "S3C2410_jsbuttons UP", 0},
	{  IRQ_EINT18,  S3C2410_GPG10,   S3C2410_GPG10_EINT18,   3, "S3C2410_jsbuttons DN", 0},
	{  IRQ_EINT19,  S3C2410_GPG11,   S3C2410_GPG11_EINT19,   4, "S3C2410_jsbuttons MID", 0},
};
*/
struct s3c2410_jsbuttons_private {
	struct device		*parent;
	struct input_dev	*dev;
	spinlock_t		lock;
	int 			count;
	int			shift;	
};

int delta_x = 5;
int delta_y = 5;
int js_sample_interval = REPEAT_DELAY;

static struct s3c2410_jsbuttons_private priv;

int js_mod_loaded = 0;

#define ENABLE_SYSFS
#ifdef ENABLE_SYSFS
static ssize_t s3c2410js_deltax_show(struct device *dev,struct device_attribute *attr,char *buf)
{
	return snprintf(buf,PAGE_SIZE,"%d \n",delta_x);
}
static ssize_t s3c2410js_deltax_store(struct device *dev,struct device_attribute *attr,const char *buf,size_t count)
{
	int temp;

        if(sscanf(buf,"%d",&temp) != 1)
	{
		return -EINVAL;
	}
        if(temp > 0)
        { 
          delta_x = temp;
        }
        else
        {
		return -EINVAL;
        }
	return strnlen(buf,count);
}
static DEVICE_ATTR(deltax,0664,s3c2410js_deltax_show,s3c2410js_deltax_store);
static ssize_t s3c2410js_deltay_show(struct device *dev,struct device_attribute *attr,char *buf)
{
	return snprintf(buf,PAGE_SIZE,"%d \n",delta_y);
}
static ssize_t s3c2410js_deltay_store(struct device *dev,struct device_attribute *attr,const char *buf,size_t count)
{
	int temp;

        if(sscanf(buf,"%d",&temp) != 1)
	{
		return -EINVAL;
	}
        if(temp > 0)
        { 
          delta_y = temp;
        }
        else
        {
		return -EINVAL;
        }
	return strnlen(buf,count);
}
static DEVICE_ATTR(deltay,0664,s3c2410js_deltay_show,s3c2410js_deltay_store);
static ssize_t s3c2410js_interval_show(struct device *dev,struct device_attribute *attr,char *buf)
{
	return snprintf(buf,PAGE_SIZE,"%d \n",js_sample_interval);
}
static ssize_t s3c2410js_interval_store(struct device *dev,struct device_attribute *attr,const char *buf,size_t count)
{

	int temp;

        if(sscanf(buf,"%d",&temp) != 1)
	{
		return -EINVAL;
	}
        if(temp > 0)
        { 
          js_sample_interval = temp;
        }
        else
        {
		return -EINVAL;
        }
	return strnlen(buf,count);
}
static DEVICE_ATTR(spleinterval,0664,s3c2410js_interval_show,s3c2410js_interval_store);

static void reg_js_attributes(struct platform_device *pdev)
{
	device_create_file(&(pdev->dev),&dev_attr_deltax);
	device_create_file(&(pdev->dev),&dev_attr_deltay);
	device_create_file(&(pdev->dev),&dev_attr_spleinterval);
}

static void remove_js_attributes(struct platform_device *pdev)
{
	device_remove_file(&(pdev->dev),&dev_attr_deltax);
	device_remove_file(&(pdev->dev),&dev_attr_deltay);
	device_remove_file(&(pdev->dev),&dev_attr_spleinterval);
}
#endif

static irqreturn_t s3c2410js_keyevent(int irq, void *dev)
{
	struct s3c2410_jsbutton *button = (struct s3c2410_jsbutton *)dev;
	int down;
        int js_keyvalue[3]={0,0,0};


	if (!button)
		return IRQ_HANDLED;
	if (!js_mod_loaded)
		return IRQ_HANDLED;
        if(timer_pending(&button->timer))
		return IRQ_HANDLED;
	down = !(s3c2410_gpio_getpin(button->pin));
	if (button->last_state == down)
		return IRQ_HANDLED;
	button->last_state = down;

        if(js_mod_loaded)
	{
	   down = !(s3c2410_gpio_getpin(button->pin));
	   button->last_state = down;
	   dprintk("%s js button %s\n",button->name, down ? "pressed" : "released");
	   switch(button->keycode)
	   {
             case 0:
              	js_keyvalue[0] = -down*delta_x;
              	break;
	     case 1:
              	js_keyvalue[0] = down*delta_x;
              	break;
             case 2:
              	js_keyvalue[1] = down*delta_y;
              	break;
             case 3:
              	js_keyvalue[1] = -down*delta_y;
              	break;
             case 4:
              	js_keyvalue[2] = down;
	}

//	   input_regs(&priv.dev, regs);
           if(down)
	   {
	   	input_report_rel(priv.dev, REL_X,js_keyvalue[0] );
	   	input_report_rel(priv.dev, REL_Y,js_keyvalue[1]);
	   	input_report_key(priv.dev, BTN_LEFT,js_keyvalue[2]);
	   	input_sync(priv.dev);
	   }
	   else
	   {
                if(button->keycode == 4)
                {
	   		input_report_rel(priv.dev, REL_X,js_keyvalue[0] );
	   		input_report_rel(priv.dev, REL_Y,js_keyvalue[1]);
	   		input_report_key(priv.dev, BTN_LEFT,js_keyvalue[2]);
	   		input_sync(priv.dev);
                }
	   }
	   if(&button->timer)
           {
              if (down)
		 mod_timer(&button->timer, jiffies + js_sample_interval);
           }
        }

	return IRQ_HANDLED;
}

static void s3c2410js_timer_callback(unsigned long data)
{
		struct s3c2410_jsbutton *button = (struct s3c2410_jsbutton *) data;
		int down;
                int js_keyvalue[3]={0,0,0};

	        down = !(s3c2410_gpio_getpin(button->pin));
	        switch(button->keycode)
	        {
                  case 0:
                    js_keyvalue[0] = down*delta_x;
                    break;
	          case 1:
                    js_keyvalue[0] = -down*delta_x;
                    break;
                  case 2:
                    js_keyvalue[1] = down*delta_y;
                    break;
                  case 3:
                    js_keyvalue[1] = -down*delta_y;
                    break;
                  case 4:
                    js_keyvalue[2] = down;
	        }


		down = !(s3c2410_gpio_getpin(button->pin));
		
		if (down != button->last_state) 
		{
			dprintk("Timer: %s js button %s\n",button->name, down ? "pressed" : "released");
	                if(button->keycode == 4)
			{
				input_report_rel(priv.dev, REL_X,js_keyvalue[0] );
	                	input_report_rel(priv.dev, REL_Y,js_keyvalue[1]);
	                	input_report_key(priv.dev, BTN_LEFT,js_keyvalue[2]);
	                	input_sync(priv.dev);
			}
			button->last_state = down;

		}
		else
		{
			if(down)
                        {
                           dprintk("Refiring Timer: %s js button %s\n",
                                    button->name, down ? "pressed" : "released");
	                   if(button->keycode < 4)
			   {
				input_report_rel(priv.dev, REL_X,js_keyvalue[0] );
	                   	input_report_rel(priv.dev, REL_Y,js_keyvalue[1]);
	                   	input_report_key(priv.dev, BTN_LEFT,js_keyvalue[2]);
	                   	input_sync(priv.dev);
			   }
			   mod_timer(&button->timer, jiffies + js_sample_interval);
			}			
		}
}

static int __init s3c2410js_probe(struct platform_device *pdev)
{
	
	struct input_dev *input_dev;
	struct s3c2410_jsbutton *s3c2410_jsbuttons;
	struct s3c2410_jsbutton *bp;
	struct s3c2410_jsbutt_mach_info *jsbutt_info;

	int i,irq,err;
	/* Initialise input stuff */
	dprintk("Initializing s3c2410-LP35-JS \n");
        jsbutt_info = pdev->dev.platform_data;
	s3c2410_jsbuttons = jsbutt_info->buttons;
	memset(&priv, 0, sizeof(struct s3c2410_jsbuttons_private));
	input_dev = input_allocate_device();
	if (!input_dev) {
		printk(KERN_ERR "Unable to allocate the input device !!\n");
		return -ENOMEM;
	}
   	priv.parent	     = &pdev->dev;
	priv.dev	     = input_dev;
//	.evbit	= { BIT(EV_KEY) | BIT(EV_REL) },
//	.keybit = { [LONG(BTN_LEFT)] = BIT(BTN_LEFT) | BIT(BTN_MIDDLE) | BIT(BTN_RIGHT) },
//	.relbit	= { BIT(REL_X) | BIT(REL_Y) },
	priv.dev->evbit[0] = BIT(EV_KEY) | BIT(EV_REL);
        priv.dev->relbit[0] =  BIT(REL_X) | BIT(REL_Y);
        priv.dev->keybit[LONG(BTN_LEFT)] = BIT(BTN_LEFT) | BIT(BTN_MIDDLE) | BIT(BTN_RIGHT);
	priv.dev->private = &priv;
	priv.dev->cdev.dev   = &pdev->dev;
	priv.dev->name = DRV_NAME;
	priv.dev->id.bustype = BUS_HOST;
	priv.dev->id.vendor = 0xDEAD;
	priv.dev->id.product = 0xBEEF;
	priv.dev->id.version = S3C2410JSBUTVERSION;

	dprintk("Initializing s3c_js button arrays \n");
        i = jsbutt_info->size;
	dprintk("Array size = %d \n",i);
	for (i = 0, bp =s3c2410_jsbuttons; i < jsbutt_info->size; bp++, i++) {
		dprintk("i = %d \n",i);
		set_bit(bp->keycode, priv.dev->keybit);
		s3c2410_gpio_pullup(bp->pin,0);	
		s3c2410_gpio_cfgpin(bp->pin, S3C2410_GPIO_SFN2);
		bp->irq = s3c2410_gpio_getirq(bp->pin);	
		init_timer(&(bp->timer));
		bp->timer.function = s3c2410js_timer_callback;
		bp->timer.data     = (unsigned long)bp;
		dprintk("js timer setup i = %d \n",i);
		irq = bp->irq;
		dprintk("i = % d requesting irq =%d \n",i,irq);
		err = request_irq (bp->irq, s3c2410js_keyevent,
			     IRQF_SAMPLE_RANDOM | IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			     bp->name, bp);
		if (err) {
			dev_err(&pdev->dev, "error requesting irq %d\n", bp->irq);
			goto err_irq;
		}
		dprintk("i = %d irq set irq = %d \n",i,irq);
//		set_irq_type(bp->irq, IRQT_BOTHEDGE);
                dprintk("done i = %d \n",i);
	}

	printk(KERN_INFO "%s successfully loaded\n", DRV_NAME);
	js_mod_loaded = 1;
	/* All went ok, so register to the input system */
	#ifdef ENABLE_SYSFS
		reg_js_attributes(pdev);
	#endif
	input_register_device(priv.dev);

	return 0;
err_irq:
	for (--i,--bp ; i >= 0; i--, bp--) {
		disable_irq(bp->irq);
		free_irq(bp->irq, bp);
	}
	input_free_device(input_dev);	
	return err;
}

static int s3c2410js_remove(struct platform_device *pdev)
{
	int i;
	struct s3c2410_jsbutton *s3c2410_jsbuttons;
	struct s3c2410_jsbutton *bp;
	struct s3c2410_jsbutt_mach_info *jsbutt_info;
        jsbutt_info = pdev->dev.platform_data;
	s3c2410_jsbuttons = jsbutt_info->buttons;

	for (i = 0,bp = s3c2410_jsbuttons; i < jsbutt_info->size; i++,bp++) {
		disable_irq(bp->irq);
		free_irq(bp->irq,bp);
	}
	#ifdef ENABLE_SYSFS
		remove_js_attributes(pdev);
	#endif
	input_unregister_device(priv.dev);

	return 0;
}


static struct platform_driver s3c2410js_driver = {
	.driver		= {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
	},
       .probe          = s3c2410js_probe,
       .remove         = s3c2410js_remove,
};


int __init s3c2410js_init(void)
{
	return platform_driver_register(&s3c2410js_driver);
}

void __exit s3c2410js_exit(void)
{
	platform_driver_unregister(&s3c2410js_driver);
}

module_init(s3c2410js_init);
module_exit(s3c2410js_exit);




